import numpy as np


def calculate_focal_length(area_mm2: float | np.ndarray, fov_degrees: float | np.ndarray) -> float | np.ndarray:
    """
    Calculate the focal length of a camera given the sensor area and field of view.

    :param area_mm2: The area of the sensor in square millimeters.
    :type area_mm2: float
    :param fov_degrees: The field of view of the camera in degrees.
    :type fov_degrees: float
    :return: The focal length of the camera in millimeters.
    :rtype: float
    """
    # Sanitise the inputs
    if np.isnan(area_mm2) or np.isnan(fov_degrees):
        raise ValueError("Input values cannot be NaN")
    if np.isinf(area_mm2) or np.isinf(fov_degrees):
        raise ValueError("Input values cannot be infinite")
    if area_mm2 <= 0:
        raise ValueError("Sensor area must be positive")
    if fov_degrees <= 0 or fov_degrees >= 180:
        raise ValueError("Field of view must be greater than 0 and less than 180")
    # Assuming a square sensor
    sensor_width_mm = np.sqrt(area_mm2)
    # Convert FOV to radians
    fov_radians = np.radians(fov_degrees)
    # Calculate tan(fov/2) to use in the focal length calculation
    tan_half_fov = np.tan(fov_radians / 2)
    if tan_half_fov == 0:
        raise ZeroDivisionError("tan(fov/2) is zero, invalid field of view")
    # Calculate focal length
    focal_length_mm = sensor_width_mm / (2 * tan_half_fov)
    return focal_length_mm



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

from ..connection import Client


class Context:
    """
    The context object contains some functions for the simulation. This can be used
    for passing data between different parts of the simulation, such as the simulation
    engine and the function library.
    """

    always_require_refresh: bool = False
    """
    Whether the context always requires a refresh, regardless of the variable data that
    has been changed.
    """

    async def get_function_library(self) -> any:
        """
        Returns the function library for the context.
        """
        raise NotImplementedError(
            "get_function_library() must be implemented by subclasses."
        )

    def get_client(self) -> Client:
        """
        Returns the client for the context.
        """
        raise NotImplementedError("get_client() must be implemented by subclasses.")

    def get_id(self) -> str:
        """
        Returns the ID of the context.
        """
        raise NotImplementedError("get_id() must be implemented by subclasses.")



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

import numpy as np


"""
[Zendir] A static library that helps with converting different
mass values between different frames of reference. These
include Center of Masses and Moment of Inertias. The frames
of reference are:
    L -> Component (Local) Frame
    B -> Body Frame (Spacecraft)
    N -> Inertial Frame
"""


def moment_of_inertia_L_to_B(
    moi_L_L: np.ndarray, dcm_LB: np.ndarray, com_B_B: np.ndarray, mass: float
) -> np.ndarray:
    """
    Converts a Moment of Inertia in the L frame to a Moment of
    Inertia in the B frame.

    :param moi_L_L: The moment of inertia in the L frame [kg m^2]
    :type moi_L_L: numpy.ndarray
    :param dcm_LB: The rotational matrix between the L frame and the B frame [-]
    :type dcm_LB: numpy.ndarray
    :param com_B_B: The center of mass in the B frame [m]
    :type com_B_B: numpy.ndarray
    :param mass: The mass of the object [kg]
    :type mass: float
    :return: The moment of inertia in the B frame [kg m^2]
    :rtype: numpy.ndarray
    """
    return inertia_point_transform(moi_L_L, dcm_LB, com_B_B, mass)


def moment_of_inertia_B_to_L(
    moi_B_B: np.ndarray, dcm_LB: np.ndarray, com_B_B: np.ndarray, mass: float
) -> np.ndarray:
    """
    Converts a Moment of Inertia in the B frame to a Moment of
    Inertia in the L frame.

    :param moi_B_B: The moment of inertia in the B frame [kg m^2]
    :type moi_B_B: numpy.ndarray
    :param dcm_LB: The rotational matrix between the L frame and the B frame [-]
    :type dcm_LB: numpy.ndarray
    :param com_B_B: The center of mass in the B frame [m]
    :type com_B_B: numpy.ndarray
    :param mass: The mass of the object [kg]
    :type mass: float
    :return: The moment of inertia in the L frame [kg m^2]
    :rtype: numpy.ndarray
    """
    # Transpose of the dcm_LB matrix to reverse the transformation
    dcm_BL = dcm_LB.T
    return inertia_inverse_point_transform(moi_B_B, dcm_BL, com_B_B, mass)


def moment_of_inertia_prime_L_to_B(
    moiPrime_L_L: np.ndarray,
    dcm_LB: np.ndarray,
    com_B_B: np.ndarray,
    comDot_B_B: np.ndarray,
    mass: float,
) -> np.ndarray:
    """
    Converts a Moment of Inertia Prime (which is a derivative) from the L
    frame to a B frame.

    :param moiPrime_L_L: The moment of inertia in the L frame [kg m^2/s]
    :type moiPrime_L_L: numpy.ndarray
    :param dcm_LB: The rotation matrix from the L to B frame [-]
    :type dcm_LB: numpy.ndarray
    :param com_B_B: The center of mass in the B frame [m]
    :type com_B_B: numpy.ndarray
    :param comDot_B_B: The center of mass change in time in the B frame [m/s]
    :type comDot_B_B: numpy.ndarray
    :param mass: The mass of the object [kg]
    :type mass: float
    :return: The moment of inertia in the B frame [kg m^2/s]
    :rtype: numpy.ndarray
    """
    return inertia_prime_point_transform(
        moiPrime_L_L, dcm_LB, com_B_B, comDot_B_B, mass
    )


def center_of_mass_L_to_B(
    com_L_L: np.ndarray, dcm_LB: np.ndarray, r_LB_B: np.ndarray
) -> np.ndarray:
    """
    Converts a Center of Mass in the L frame to a Center of Mass
    in the B frame.

    :param com_L_L: The center of mass in the L frame [m]
    :type com_L_L: numpy.ndarray
    :param dcm_LB: The rotation matrix from the L to B frames [-]
    :type dcm_LB: numpy.ndarray
    :param r_LB_B: The location difference between the L and B frames [m]
    :type r_LB_B: numpy.ndarray
    :return: The center of mass in the B frame [m]
    :rtype: numpy.ndarray
    """
    r_LmL_L = com_L_L
    r_LmL_B = np.dot(dcm_LB.T, r_LmL_L)
    r_LmB_B = r_LmL_B + r_LB_B
    return r_LmB_B


def center_of_mass_B_to_L(
    com_B_B: np.ndarray, dcm_LB: np.ndarray, r_LB_B: np.ndarray
) -> np.ndarray:
    """
    Converts a Center of Mass in the B frame to a Center of Mass
    in the L frame.

    :param com_B_B: The center of mass in the B frame [m]
    :type com_B_B: numpy.ndarray
    :param dcm_LB: The rotation matrix from the L to B frames [-]
    :type dcm_LB: numpy.ndarray
    :param r_LB_B: The location difference between the L and B frames [m]
    :type r_LB_B: numpy.ndarray
    :return: The center of mass in the L frame [m]
    :rtype: numpy.ndarray
    """
    r_LmB_B = com_B_B
    r_LmL_B = r_LmB_B - r_LB_B
    r_LmL_L = np.dot(dcm_LB, r_LmL_B)
    return r_LmL_L


def tensor_matrix(tensor: np.ndarray, dcm: np.ndarray) -> np.ndarray:
    """
    Transforms a given tensor using a directional cosine matrix (DCM).

    :param tensor: The input tensor to be transformed [3x3 matrix]
    :type tensor: numpy.ndarray
    :param dcm: The directional cosine matrix used for the transformation [-]
    :type dcm: numpy.ndarray
    :return: The transformed tensor [3x3 matrix]
    :rtype: numpy.ndarray
    """
    return np.dot(dcm.T, np.dot(tensor, dcm))


def tensor_transform(tensor: np.ndarray, dcm: np.ndarray) -> np.ndarray:
    """
    Transform a rank-2 tensor to be represented in a new frame.

    :param tensor: A rank-2 tensor in the original frame [3x3 matrix]
    :type tensor: numpy.ndarray
    :param dcm: The DCM for the frame transformation [-]
    :type dcm: numpy.ndarray
    :return: The transformed tensor in the new frame [3x3 matrix]
    :rtype: numpy.ndarray
    """
    return dcm.T @ tensor @ dcm


def inertia_point_transform(
    inertia: np.ndarray, dcm: np.ndarray, position: np.ndarray, mass: float
) -> np.ndarray:
    """
    Transform an inertia tensor from the center-of-mass frame to a new frame.

    :param inertia: The inertia tensor in the original frame [3x3 matrix]
    :type inertia: numpy.ndarray
    :param dcm: The DCM of the transformation [-]
    :type dcm: numpy.ndarray
    :param position: The position of the new frame relative to the original frame [m]
    :type position: numpy.ndarray
    :param mass: The mass of the object [kg]
    :type mass: float
    :return: The transformed inertia tensor in the new frame [3x3 matrix]
    :rtype: numpy.ndarray
    """
    i_parallel = tensor_matrix(inertia, dcm)
    r_tilde = skew_matrix(position)
    return i_parallel - mass * np.dot(r_tilde, r_tilde)


def inertia_inverse_point_transform(
    inertia: np.ndarray, dcm: np.ndarray, position: np.ndarray, mass: float
) -> np.ndarray:
    """
    Transform an inertia tensor from a new frame back to the center-of-mass frame.

    :param inertia: The inertia tensor in the new frame [3x3 matrix]
    :type inertia: numpy.ndarray
    :param dcm: The DCM of the transformation [-]
    :type dcm: numpy.ndarray
    :param position: The position of the new frame relative to the original frame [m]
    :type position: numpy.ndarray
    :param mass: The mass of the object [kg]
    :type mass: float
    :return: The transformed inertia tensor in the original frame [3x3 matrix]
    :rtype: numpy.ndarray
    """
    r_tilde = skew_matrix(position)
    i_parallel = inertia + mass * np.dot(r_tilde, r_tilde)
    return tensor_transform(i_parallel, dcm)


def inertia_prime_point_transform(
    inertia_prime: np.ndarray,
    dcm: np.ndarray,
    position: np.ndarray,
    position_dot: np.ndarray,
    mass: float,
) -> np.ndarray:
    """
    Transform an InertiaPrime tensor from the center-of-mass frame to a new frame.

    :param inertia_prime: The InertiaPrime tensor in the original frame [3x3 matrix]
    :type inertia_prime: numpy.ndarray
    :param dcm: The DCM of the transformation [-]
    :type dcm: numpy.ndarray
    :param position: The position of the new frame relative to the original frame [m]
    :type position: numpy.ndarray
    :param position_dot: The rate of change of the position of the center of mass [m/s]
    :type position_dot: numpy.ndarray
    :param mass: The mass of the object [kg]
    :type mass: float
    :return: The transformed InertiaPrime tensor in the new frame [3x3 matrix]
    :rtype: numpy.ndarray
    """
    # Transform InertiaPrime to a frame parallel to the target frame
    i_parallel = tensor_matrix(inertia_prime, dcm)

    # Compute skew matrices for the position and position rate (position_dot)
    r_tilde = skew_matrix(position)
    r_tilde_prime = skew_matrix(position_dot)

    # Apply the parallel axis theorem to the InertiaPrime tensor
    return i_parallel - mass * (
        np.dot(r_tilde_prime, r_tilde) + np.dot(r_tilde, r_tilde_prime)
    )


def skew_matrix(vector: np.ndarray) -> np.ndarray:
    """
    Generates a skew-symmetric matrix from a 3-element vector.

    :param vector: The input vector [m]
    :type vector: numpy.ndarray
    :return: The skew-symmetric matrix [3x3 matrix]
    :rtype: numpy.ndarray
    """
    return np.array(
        [
            [0, -vector[2], vector[1]],
            [vector[2], 0, -vector[0]],
            [-vector[1], vector[0], 0],
        ]
    )



# Define the version of the package
__version__ = "1.4.0.0"

# Import the standard utilities
from .utils import ZendirException, printer, helper

# Import the standard classes that are commonly used
from .connection import Client
from .simulation import *

# Include the runner utilities
from .utils import runner



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

from __future__ import annotations
import json, os
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
from .. import ZendirException


class SimulationData:
    """
    The DataFrame class is able to store a series of parameters and data associated
    with a particular simulation instance. This class is a pure data class and is
    not able to invoke any methods. It is able to store the data in a structured
    format and is able to be exported to a JSON format.
    """

    __raw: dict = {}
    """Defines the raw data that is fetched from the API."""

    type: str = ""
    """Defines the object type that was used for this SimulationData."""

    id: str = ""
    """Defines the unique GUID identifier of the object. This needs to be in the correct GUID format."""

    fields: list = []
    """Defines the fields that are associated with the SimulationData."""

    data: list = []
    """Defines the data that is associated with the SimulationData."""

    def __init__(self, data: dict) -> None:
        """
        Initialises the SimulationData with the data that is fetched from the API. This must be
        created using the JSON data that is fetched from the API and is able to be exported
        back to the API in the same format.

        :param data:    The data that is fetched from the API
        :type data:     dict
        """

        # Reset the data
        self.id = ""
        self.type = ""
        self.fields = []
        self.data = []
        self.__raw = data

        # Fetch the header values
        if "ID" in data:
            self.id = data["ID"]
        if "Type" in data:
            self.type = data["Type"]
        if "Data" not in data:
            raise ZendirException(
                "Data not found. Please ensure the queried data name is correct and has been tracked."
            )

        # Data should be an array of arrays
        data = list(data["Data"])
        if len(data) == 0:
            raise ZendirException(
                "SimulationData Data is empty. Please ensure the queried data name is correct and has been tracked."
            )
        self.fields = data[0]

        # Loop through the data
        for i in range(1, len(data)):
            self.data.append(data[i])

    @classmethod
    def load(cls, path: str) -> SimulationData:
        """
        Loads a SimulationData from a file path and returns the SimulationData object. This is a
        class method and does not require an instance to be created.

        :param path:    The path to the file that contains the SimulationData
        :type path:     str

        :returns:       The SimulationData object that is loaded from the file
        :rtype:         SimulationData
        """

        # Check if the path is not valid and throw exception
        if not path or not os.path.exists(path):
            raise ZendirException("Invalid path provided to load SimulationData.")
        return SimulationData(json.loads(path))

    def get_times(self) -> np.ndarray:
        """
        Returns the time values that are associated with the SimulationData. This is a helper
        function that is able to fetch the time values from the SimulationData, where each
        value is in seconds.

        :returns:   The time values that are associated with the SimulationData in seconds
        :rtype:     np.ndarray
        """

        # Ensure that the time is casted as a float as a numpy arra
        times = [float(x) for x in self.get_values("Time")]

        # Convert and return as a numpy array
        return np.array(times)

    def get_values(self, parameter: str) -> np.ndarray:
        """
        Returns the values that are associated with a particular parameter in the SimulationData.
        This is a helper function that is able to fetch the values associated with a particular
        parameter in the SimulationData.

        :param parameter:   The parameter to fetch the values for
        :type parameter:    str

        :returns:           The values that are associated with the parameter
        :rtype:             np.ndarray
        """

        # Check if the parameter does not exist
        if parameter not in self.fields:
            raise ZendirException(
                f"Parameter '{parameter}' not found in SimulationData."
            )

        # Fetch the index and data associated with it
        index = self.fields.index(parameter)
        values: list = [row[index] for row in self.data]
        return np.array(values)

    def plot(self, title="Simulation Data", params=None) -> None:
        """
        Creates a quick plot of the SimulationData data. This will plot the data in a simple
        line plot format and will display the data in a single plot or multiple plots
        depending on the parameters provided. The following formats for the parameters
        will work:

        params = "MyParam"                    -> will plot 'MyParam' against Time in a single plot
        params = ["MyParam1", "MyParam2"]     -> will plot 'MyParam1' and 'MyParam2' against Time in a single plot
        params = [["MyParam1"], ["MyParam2"]] -> will plot 'MyParam1' and 'MyParam2' against Time in separate plots

        :param title:   The title of the plot
        :type title:    str
        :param params:  The parameters to plot
        :type params:   str or list
        """

        # Ensure params is initialized to an empty list if not provided
        if params is None:
            params = []

        # Check if params is a single string (single parameter case)
        if isinstance(params, str):
            params = [params]  # Convert to a list to handle uniformly

        # Check if params is a list of lists (multiple plots case)
        if all(isinstance(item, list) for item in params):
            _, axs = plt.subplots(len(params), 1, figsize=(8, len(params) * 4))
            # Ensure axs is always a list (even if there's only one subplot)
            if len(params) == 1:
                axs = [axs]

            for ax, param_list in zip(axs, params):
                for parameter in param_list:
                    ax.plot(
                        self.get_times(), self.get_values(parameter), label=parameter
                    )
                ax.set(xlabel="Time [s]", ylabel="Value", title=title)
                ax.grid()
                ax.legend()
        else:
            # Handle the case where params is a flat list
            _, ax = plt.subplots()
            for parameter in params:
                ax.plot(self.get_times(), self.get_values(parameter), label=parameter)
            ax.set(xlabel="Time [s]", ylabel="Value", title=title)
            ax.grid()
            ax.legend()

        # Display the plots on the screen
        plt.tight_layout()
        plt.show()

    def export(self) -> dict:
        """
        Exports the SimulationData to a dictionary format that is able to be exported to the API.
        This will return the SimulationData in the same format that is fetched from the API.

        :returns:   The SimulationData in a dictionary format
        :rtype:     dict
        """
        return self.__raw

    def save(self, path: str) -> None:
        """
        Saves the SimulationData to a file path in a JSON format. This will save the SimulationData
        in a JSON format that is able to be loaded back into the SimulationData class.

        :param path:    The path to save the SimulationData to
        :type path:     str
        """
        with open(path, "w") as f:
            f.write(self.__str__())

    def to_dataframe(self) -> pd.DataFrame:
        """
        Converts the SimulationData to a pandas DataFrame. This will convert the SimulationData
        to a pandas DataFrame and will return the DataFrame object.

        :returns:   The SimulationData in a pandas DataFrame format
        :rtype:     pd.DataFrame
        """
        # Create a dictionary of the data
        data = {}
        for row in self.data:
            for i, field in enumerate(self.fields):
                # If any of the row values are a lists, then break the list into separate columns
                if isinstance(row[i], list):
                    for j, value in enumerate(row[i]):
                        # Check if the key exists in the dictionary, if not create it
                        if f"{field}_{j}" not in data:
                            data[f"{field}_{j}"] = []
                        data[f"{field}_{j}"].append(value)
                else:
                    # Check if the key exists in the dictionary, if not create it
                    if field not in data:
                        data[field] = []
                    data[field].append(row[i])
        return pd.DataFrame(data)

    def __str__(self) -> str:
        """
        Converts the SimulationData to a string format that is able to be exported to a JSON format.

        :returns:   The SimulationData in a string format
        :rtype:     str
        """
        return json.dumps(self.export(), indent=4)



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

import numpy as np


def to_dcm(
    right: np.ndarray = None, forward: np.ndarray = None, up: np.ndarray = None
) -> np.ndarray:
    """
    Converts the DCM to a new DCM using the provided vectors.

    :param right: A unit vector representing the right axis.
    :param forward: A unit vector representing the forward axis.
    :param up: A unit vector representing the up axis.
    :return: The new DCM.
    """
    # Normalize the input vectors
    if right is not None and forward is not None and up is not None:
        axis_u = up / np.linalg.norm(up)
        axis_r = right / np.linalg.norm(right)
        axis_f = forward / np.linalg.norm(forward)

        # Assign matrix components
        M = np.array(
            [
                [axis_r[0], axis_f[0], axis_u[0]],
                [axis_r[1], axis_f[1], axis_u[1]],
                [axis_r[2], axis_f[2], axis_u[2]],
            ]
        )
    else:
        # Default to identity matrix if no vectors are provided
        M = np.eye(3)
    return M


def up_axis_to_dcm(up: np.ndarray) -> np.ndarray:
    """
    Returns a DCM matrix that aligns the up axis with the given direction.

    :param up: The up axis to align with.
    :return: The DCM aligned to the up axis.
    """
    right_vec = np.array([1, 0, 0])
    forward_vec = np.array([0, 1, 0])
    up_vec = np.array([0, 0, 1])

    # Skip if a zero vector
    if np.linalg.norm(up) == 0:
        raise ValueError("Zero vector provided for up axis.")

    # Check if vectors are aligned
    check = np.dot(right_vec, up)

    # Aligned with up axis
    if check >= 1 - 1e-15:
        return to_dcm(-up_vec, forward_vec, right_vec)

    # Anti-aligned with up axis (down axis)
    if check <= -1 + 1e-15:
        return to_dcm(up_vec, forward_vec, -right_vec)

    # Calculate the axis
    up_norm = up / np.linalg.norm(up)
    forward = np.cross(up_norm, right_vec)
    right = np.cross(forward, up)

    # Return the matrix
    return to_dcm(right, forward, up)


def mrp_to_dcm(mrp: np.ndarray) -> np.ndarray:
    """
    Converts a Modified Rodrigues Parameter (MRP) to a Direction Cosine Matrix (DCM).

    :param mrp: The MRP to convert.
    :return: The DCM representation of the MRP.
    """

    q1 = mrp[0]
    q2 = mrp[1]
    q3 = mrp[2]

    d1 = np.dot(mrp, mrp)
    s = 1 - d1
    d = (1 + d1) * (1 + d1)
    c = np.identity(3)
    c[0, 0] = 4 * (2 * q1 * q1 - d1) + s * s
    c[0, 1] = 8 * q1 * q2 + 4 * q3 * s
    c[0, 2] = 8 * q1 * q3 - 4 * q2 * s
    c[1, 0] = 8 * q2 * q1 - 4 * q3 * s
    c[1, 1] = 4 * (2 * q2 * q2 - d1) + s * s
    c[1, 2] = 8 * q2 * q3 + 4 * q1 * s
    c[2, 0] = 8 * q3 * q1 + 4 * q2 * s
    c[2, 1] = 8 * q3 * q2 - 4 * q1 * s
    c[2, 2] = 4 * (2 * q3 * q3 - d1) + s * s
    c *= 1 / d

    return c


def euler2(angle_rad: float) -> np.ndarray:
    """
    Create a rotation matrix for a rotation about the Y-axis.

    :param angle_rad: Rotation angle in radians.
    :type angle_rad: float
    :return: 3x3 rotation matrix about the Y-axis.
    :rtype: numpy.ndarray
    """
    cos_x = np.cos(angle_rad)
    sin_x = np.sin(angle_rad)
    rotation_matrix = np.array(
        [[cos_x, 0.0, sin_x], [0.0, 1.0, 0.0], [-sin_x, 0.0, cos_x]]
    )
    return rotation_matrix


def euler3(angle_rad: float) -> np.ndarray:
    """
    Create a rotation matrix for a rotation about the Z-axis.

    :param angle_rad: Rotation angle in radians.
    :type angle_rad: float
    :return: 3x3 rotation matrix about the Z-axis.
    :rtype: numpy.ndarray
    """
    cos_x = np.cos(angle_rad)
    sin_x = np.sin(angle_rad)
    rotation_matrix = np.array(
        [[cos_x, -sin_x, 0.0], [sin_x, cos_x, 0.0], [0.0, 0.0, 1.0]]
    )
    return rotation_matrix


def rotate_dcm(dcm: np.ndarray, axis: np.ndarray, angle: float) -> np.ndarray:
    """
    Rotates a DCM matrix about an axis by a given angle. The DCM is assumed to be
    a 3x3 matrix and the axis is a 3x1 vector. The angle is measured in degrees.

    :param dcm: The DCM matrix to rotate, assumed to be a 3x3 matrix.
    :type dcm: numpy.ndarray
    :param axis: The axis to rotate about, assumed to be a 3x1 vector.
    :type axis: numpy.ndarray
    :param angle: The angle of rotation in degrees.
    :type angle: float

    :return: The rotated DCM matrix.
    :rtype: numpy.ndarray
    """

    # If a list is used, turn into a numpy array
    if isinstance(dcm, list):
        dcm = np.array(dcm)
    if isinstance(axis, list):
        axis = np.array(axis)

    # Validate the data
    if dcm.shape != (3, 3):
        raise ValueError("DCM matrix must be a 3x3 matrix.")
    if axis.shape != (3,):
        raise ValueError("Axis must be a 3x1 vector.")

    # Convert the angle to radians
    angle_rad = np.radians(angle)

    # Compute the rotation matrix
    c = np.cos(angle_rad)
    s = np.sin(angle_rad)
    t = 1 - c
    x = axis[0]
    y = axis[1]
    z = axis[2]

    # Compute the rotation matrix
    rotation_matrix = np.array(
        [
            [t * x * x + c, t * x * y - s * z, t * x * z + s * y],
            [t * x * y + s * z, t * y * y + c, t * y * z - s * x],
            [t * x * z - s * y, t * y * z + s * x, t * z * z + c],
        ]
    )

    # Return the rotated DCM
    return np.dot(rotation_matrix, dcm)



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

from .instance import Instance
from .context import Context


class Message(Instance):
    """
    The Message class is able to store a series of parameters and data
    associated with the particular message type. This class is a pure
    data class and is not able to invoke any methods.
    """

    def __init__(self, context: Context, id: str, type: str = None) -> None:
        """
        Initialises the message with a context and a
        unique GUID identifier for the message.

        :param context:     The context object that is used for the API
        :type context:      Context
        :param id:          The unique identifier for the Entity in a GUID format
        :type id:           str
        :param type:        The type of the system, if applicable
        :type type:         str
        """

        super().__init__(context, id, type)



from .behaviour import Behaviour
from .instance import Instance
from .message import Message
from .model import Model
from .object import Object
from .simulation import Simulation
from .system import System


# ---------------------------------------------------------------------------------------------------------------------------- #
# Copyright 2025 (c) Zendir, Pty Ltd. All Rights Reserved
# See the 'LICENSE' file at the root of this package
# ---------------------------------------------------------------------------------------------------------------------------- #
import aiohttp, json, typing
# ---------------------------------------------------------------------------------------------------------------------------- #

async def rqst(method: str, url: str, data: typing.Any = None, headers: dict = None) -> typing.Any:

    """
    Sends a HTTP request with the specified request and url and returns a response.
    """

    # parse HTTP request headers
    if headers and not isinstance(headers, dict):
        raise Exception("invalid argument 'headers'")

    # parse HTTP request content body
    content = None
    if data:
        if headers is None: headers = {}
        if isinstance(data, str):
            content = data.encode()
            headers["Content-Type"]   = "text/plain"
            headers["Content-Length"] = str(len(content))
        elif isinstance(data, (dict, list)):
            content = json.dumps(data).encode()
            headers["Content-Type"]   = "application/json"
            headers["Content-Length"] = str(len(content))
        else:
            raise Exception("invalid argument 'body'")

    # send HTTP request and wait for response
    results = {}
    async with aiohttp.ClientSession() as session:
        async with session.request(method, url, data=content, headers=headers) as response:
            results["body"]    = await response.read()
            results["status"]  = response.status
            results["headers"] = dict(response.headers)
    if results["status"] != 200:
        raise Exception(results["body"].decode() if results["body"] else "")
    if results["body"]:
        if "Content-Type" not in results["headers"]:
            raise Exception("missing 'Content-Type'")
        match results["headers"]["Content-Type"]:
            case "text/plain":
                return results["body"].decode()
            case "application/json":
                return json.loads(results["body"].decode())
    return None

# ---------------------------------------------------------------------------------------------------------------------------- #


#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

from ..utils import ZendirException, helper
from .context import Context


class Instance:
    """
    The instance class is a base class that is used to define any simulation object
    that exists within the simulation. It is able to fetch and set data on the API,
    invoke methods and can be tracked by the API. This class is used to define the
    base functionality for the other types, but should not be created by the user
    itself.
    """

    __data: dict = None
    """Defines the data dictionary that is fetched from the API."""

    __type: str = None
    """Defines the type of the object that is fetched from the API."""

    _refresh_cache: bool = True
    """Defines whether the cache needs to be refreshed or not."""

    _ignore_refresh_override: bool = False
    """Defines an override for when to ignore the required refresh, so that when fetching data, there is no recursion."""

    _context: Context = None
    """Defines the context that is used to access the API."""

    __id: str = None
    """Defines the unique GUID identifier of the object. This needs to be in the correct GUID format."""

    def __init__(self, context: Context, id: str, type: str = None) -> None:
        """
        Initialises the instance with the context and the ID of the object.

        :param context:         The context used to access the API
        :type context:          Context
        :param id:              The GUID ID of the object
        :type id:               str
        :param type:            The type of the object, if known
        :type type:             str
        """

        # Check if the context and ID are valid
        if not context:
            raise ZendirException("Failed to create instance due to invalid context.")
        if not context.get_client():
            raise ZendirException("Failed to create instance due to invalid client.")
        if not helper.is_valid_guid(id):
            raise ZendirException(
                f"Failed to create instance with an invalid ID '{id}'."
            )

        # Set the values and default data
        self.__id = id
        self._context = context
        self.__data = None
        self.__type = type
        self._refresh_cache = True

    async def _get_data(self) -> None:
        """
        This function is used to fetch the data from the API and store it in the
        object. This is used to ensure that the data is fetched correctly and is
        up to date.
        """

        # Skip if the cache does not need to be refreshed
        if not self._refresh_cache:
            return

        # Fetch all data and then set the cache to false
        self.__data = await self._context.get_client().get(
            f"{self.get_id()}/get", id=self._context.get_id()
        )
        self._refresh_cache = (
            self._context.always_require_refresh and not self._ignore_refresh_override
        )

        # Loop through the data and deserialize it
        for key in self.__data:
            self.__data[key] = helper.deserialize(self.__data[key])

    def _reset_refresh_cache(self) -> None:
        """
        This function is used to set the flag for refreshing the cache to true.
        It can be overriden by a base class if additional functionality is required.
        """

        self._refresh_cache = True

    def get_id(self) -> str:
        """
        Returns the unique GUID identifier of the object. This is used to identify
        the object in the API and is used to fetch the data from the API.

        :returns:   The unique GUID identifier of the object
        :rtype:     str
        """

        # Return the ID of the object
        return self.__id

    async def get_type(self) -> str:
        """
        Returns the type of the object that is fetched from the API. This is the
        full type, including the namespace and the class name. If the type has not
        been fetched, it will fetch it from the API.

        :returns:   The type of the object
        :rtype:     str
        """

        # If the type is not provided, or empty, then fetch it from the API
        if not self.__type or self.__type == "":

            # Fetch the function library from the context
            function_library: Instance = await self._context.get_function_library()
            if not function_library:
                raise ZendirException(
                    "Failed to get type of instance due to missing function library."
                )

            # Fetch the type from the API
            self.__type = await function_library.invoke("GetObjectType", self.get_id())

            # If the type is still None, raise an exception
            if not self.__type or self.__type == "":
                raise ZendirException(
                    f"Failed to get type of instance with ID '{self.get_id()}'."
                )

            # Set the type to only be the last part of the "." separated string
            self.__type = self.__type.split(".")[-1]

        # If there is no type, find it
        return self.__type

    def get_instance_with_id(self, id: str, recurse: bool = False) -> "Instance":
        """
        Returns the instance that is attached to this instance with the specified ID. If the
        instance does not exist, None will be returned. This may be a message or a child instance
        of some kind.

        :param id:  The ID of the instance to fetch
        :type id:   str
        :param recurse:  Whether to look down the chain of children to find the instance
        :type recurse:   bool

        :returns:   The instance that is attached to the object with the specified ID
        :rtype:     Instance
        """

        # To be overriden by the base class, this is just a placeholder
        return None

    async def get(self, param: str) -> any:
        """
        Fetches a parameter from the object that is stored in the API. This will
        return the parameter if it exists and will raise an exception if it does
        not exist. If the data has not been fetched, it will fetch the data from
        the API.

        :param param:   The parameter to fetch from the object
        :type param:    str

        :returns:       The parameter value from the object
        :rtype:         any
        """

        # Ensures that the data is fetched correctly
        await self._get_data()

        # Check if the param is in the data
        if param not in self.__data:
            raise ZendirException(
                f"Failed to get parameter '{param}' in object '{self.get_id()}' of type '{await self.get_type()}'."
            )
        return self.__data[param]

    async def get_all(self) -> dict:
        """
        Returns all of the data that is stored in the object. This will return
        the full dictionary of the object and will fetch the data if it has not
        been fetched yet.

        :returns:   The full dictionary of the object
        :rtype:     dict
        """

        # Ensures that the data is fetched correctly
        await self._get_data()
        return self.__data

    async def set(self, **kwargs) -> None:
        """
        Sets the parameters of the object that is stored in the API. This will
        take in a dictionary of parameters and values and will update the object
        with the new values. This will raise an exception if the data is invalid.

        :param kwargs:  The parameters and values to set on the object
        :type kwargs:   dict
        """

        # For each of the key values, serialize the data
        for key in kwargs:
            kwargs[key] = helper.serialize(kwargs[key])

        # Call the method on the client
        await self._context.get_client().post(
            f"{self.get_id()}/set", data=kwargs, id=self._context.get_id()
        )

        # Ensure that the cache is refreshed for the next get
        self._reset_refresh_cache()

    async def invoke(self, function: str, *args) -> any:
        """
        Invoke a function on the object that is stored in the API. This will take
        in the function name and the arguments to pass to the function. This will
        return the response from the function call, if there is a response.

        :param function:    The function name to invoke on the object
        :type function:     str
        :param args:        The arguments to pass to the function
        :type args:         any

        :returns:           The response from the function call
        :rtype:             any
        """

        # For each of the arguments, serialize the data
        args = list(args)
        for i in range(len(args)):
            args[i] = helper.serialize(args[i])

        # Insert the function name as the first argument
        args.insert(0, function)

        # Create the response from the invoke request and get the return data
        response = await self._context.get_client().post(
            f"{self.get_id()}/ivk", args, id=self._context.get_id()
        )
        response = helper.deserialize(response)

        # Update the flag for needing to get values
        self._reset_refresh_cache()
        return response

    def __str__(self) -> str:
        """
        The automated cast of this object to a string will only return the
        GUID identifier of the object.

        :returns:   The formatted ID from the object
        :rtype:     str
        """

        # Return the ID of the object
        return self.get_id()



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

import json


def create_command(trigger: dict, type: str, parameters: dict, priority: int = 0) -> dict:
    """
    Creates a command with a list of arguments as the parameters. This will create the JSON
    object with the list of parameters required.

    :param trigger: The trigger for the command.
    :type trigger: dict
    :param command: The command to execute.
    :type command: str
    :param parameters: The parameter dictionary containing keys and values for the command.
    :type parameters: dict
    :param priority: The priority of the command, default is 0.
    :type priority: int, optional
    :return: The JSON command that can be used.
    :rtype: dict
    """
    # Initialize an empty command dictionary
    cmd = {}

    # Set the 'command' field of the JSON object
    cmd["Trigger"] = trigger
    cmd["Type"] = type

    # Initialize an empty dictionary for parameters
    param_obj = {}

    # Iterate through the parameters and add them to the dictionary
    for key, value in parameters.items():
        param_obj[key] = value

    # Set the 'parameters' field of the JSON object
    cmd["Parameters"] = param_obj

    # Set the 'priority' field of the JSON object
    cmd["Priority"] = priority

    # Return the complete command object (JSON)
    return cmd


def create_guidance_start_command(
    trigger: dict,
    navigation: str = "Simple",
    pointing: str = "Inertial",
    controller: str = "MRP",
    mapping: str = "ReactionWheels",
    priority: int = 0,
) -> dict:
    """
    Creates a guidance command for the spacecraft operation computer to execute
    once loaded into the system at a particular time.

    :param trigger: The trigger for the command.
    :type trigger: dict
    :param navigation: The navigation mode to set (default is "Simple").
    :type navigation: str, optional
    :param pointing: The pointing mode to set (default is "Inertial").
    :type pointing: str, optional
    :param controller: The controller mode to set (default is "MRP").
    :type controller: str, optional
    :param mapping: The mapping mode to set (default is "ReactionWheels").
    :type mapping: str, optional
    :param priority: The priority of the command, default is 0.
    :type priority: int, optional
    :return: The appropriate command JSON to be executed.
    :rtype: dict
    """
    # Create a dictionary of arguments with navigation, pointing, controller, and mapping types
    args = {
        "Navigation": navigation,
        "Pointing": pointing,
        "Controller": controller,
        "Mapping": mapping,
    }

    # Return the generated command by calling create_command with the arguments
    return create_command(trigger, "GuidanceStart", args, priority)

def create_guidance_configure_command(
    trigger: dict,
    parameters: dict,
    priority: int = 0,
) -> dict:
    """
    Creates a guidance configure command for the spacecraft operation computer to execute
    once loaded into the system.

    :param trigger: The trigger for the command.
    :type trigger: dict
    :param parameters: The parameter dictionary containing keys and values for the command.
    :type parameters: dict
    :param priority: The priority of the command, default is 0.
    :type priority: int, optional
    :return: The appropriate command JSON to be executed.
    :rtype: dict
    """
    # Return the generated command by calling create_command with the arguments
    return create_command(trigger, "GuidanceConfigure", parameters, priority)

def create_event_trigger(type: str, repeat: bool = False, is_done: bool = False) -> dict:
    """
    Creates a event trigger command for the spacecraft operation computer to execute
    once loaded into the system.

    :param type: The type of event trigger to create.
    :type type: str
    :param repeat: Whether the event trigger should repeat, default is False.
    :type repeat: bool, optional
    :param is_done: Whether the event trigger is done, default is False.
    :type is_done: bool, optional
    :return: An event trigger for a command.
    :rtype: dict
    """
    # Create a dictionary of arguments with type, repeat, and is_done
    args = {
        "Type": type,
        "Repeat": repeat,
        "IsDone": is_done,
    }

    return args

def create_time_event_trigger(time: float, interval: float = 0.0, repeat: bool = False, is_done: bool = False) -> dict:
    """
    Creates a time event trigger command for the spacecraft operation computer to execute
    once loaded into the system. It will repeat at the interval specified.

    :param time: The time [s] at which the event trigger is executed.
    :type time: float
    :param interval: The interval [s] at which the event trigger is executed.
    :type interval: float
    :param repeat: Whether the event trigger should repeat, default is False.
    :type repeat: bool, optional
    :param is_done: Whether the event trigger is done, default is False.
    :type is_done: bool, optional
    :return: An event trigger for a command based on time.
    :rtype: dict
    """
    # Create a dictionary of arguments with time and interval
    args = create_event_trigger("Time", repeat, is_done)
    args["Time"] = time
    args["Interval"] = interval

    return args

def create_parameter_event_trigger(object_id: str, parameter_name: str, value: float, operator: str, repeat: bool = False, is_done: bool = False) -> dict:
    """
    Creates a parameter event trigger command for the spacecraft operation computer to execute
    once loaded into the system.

    :param object_id: The ID of the object to trigger the event on.
    :type object_id: str
    :param parameter_name: The name of the parameter to trigger the event on.
    :type parameter_name: str
    :param value: The value of the parameter to trigger the event on.
    :type value: float
    :param operator: The operator to use for the event trigger.
    :type operator: str
    :param repeat: Whether the event trigger should repeat, default is False.
    :type repeat: bool, optional
    :param is_done: Whether the event trigger is done, default is False.
    :type is_done: bool, optional
    :return: An event trigger for a command based on a parameter.
    :rtype: dict
    """
    args = create_event_trigger("Parameter", repeat, is_done)
    args["ObjectID"] = object_id
    args["ParameterName"] = parameter_name
    args["Value"] = value
    args["Operator"] = operator

    return args


#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

from abc import ABC, abstractmethod
from collections.abc import Container
from typing import Dict, Optional, Union, List

import numpy as np

from ..maths.astro import (
    classical_to_vector_elements,
    mean_to_osculating_elements,
    get_planet_property,
    argument_of_latitude,
)


def state_vectors(method):
    """
    Decorator to convert a collection of orbital elements into their equivalent inertial state vectors
    :param method: the method to be decorated
    :return: the decorated method
    """

    def wrapper(self, *args, **kwargs):
        orbital_elements = method(self, *args, **kwargs)
        result = dict()
        # use dictionary comprehension to efficiently convert the state vectors to classical orbital elements
        for i, elements in orbital_elements.items():
            vectors = classical_to_vector_elements(
                semi_major_axis=elements["semi_major_axis"],
                eccentricity=elements["eccentricity"],
                inclination=elements["inclination"],
                right_ascension=elements["right_ascension"],
                argument_of_periapsis=elements["argument_of_periapsis"],
                true_anomaly=elements["true_anomaly"],
                *args,
                **kwargs,
            )
            result[i] = {
                "r_bn_n": np.array(vectors[0]),
                "v_bn_n": np.array(vectors[1]),
            }
        return result

    return wrapper


def classical_elements_mean(mean_to_osculating=False):
    """
    Decorator to calculate the mean classical orbital elements from osculating elements
    :param method: the method to be decorated
    :return: the decorated method
    """

    def decorator(method):
        def wrapper(self, *args, **kwargs):
            orbital_elements = method(self, *args, **kwargs)
            planet = kwargs.get("planet", "earth")
            req = get_planet_property(planet, "REQ")
            j2 = get_planet_property(planet, "J2")
            # define the keys that will be used to access the orbital elements
            keys = (
                "semi_major_axis",
                "eccentricity",
                "inclination",
                "right_ascension",
                "argument_of_periapsis",
                "true_anomaly",
            )
            # use dictionary comprehension to efficiently convert between mean and osculating orbital elements or
            #   vice versa
            return {
                # use dictionary comprehension to map the keys to their respective orbital elements
                i: {
                    keys[j]: el
                    for j, el in enumerate(
                        mean_to_osculating_elements(
                            req=req,
                            j2=j2,
                            semi_major_axis=elements["semi_major_axis"],
                            eccentricity=elements["eccentricity"],
                            inclination=elements["inclination"],
                            right_ascension=elements["right_ascension"],
                            argument_of_periapsis=elements["argument_of_periapsis"],
                            true_anomaly=elements["true_anomaly"],
                            mean_to_osculating=mean_to_osculating,
                        )
                    )
                }
                for i, elements in orbital_elements.items()
            }

        return wrapper

    return decorator


class Constellation(ABC):

    _spacecraft: Dict[int, Dict] = None

    # the number of satellites in the constellation. There must be at least one spacecraft in the constellation
    _num_satellites: int = None

    # the semi-major axis of every spacecraft in the co-planar constellation
    _semi_major_axis: float = None
    # the eccentricity of every spacecraft in the co-planar constellation
    _eccentricity: float = 0.0
    # the inclination of every spacecraft in the co-planar constellation
    _inclination: float = None
    # the right ascension of every spacecraft in the co-planar constellation
    _right_ascension: float = 0.0
    # the argument of periapsis of every spacecraft in the co-planar constellation
    _argument_of_periapsis: float = 0.0
    # the reference true anomaly offset of every spacecraft in the co-planar constellation
    _true_anomaly_offset: float = 0.0

    @property
    def num_satellites(self) -> int:
        """
        get the number of satellites in the constellation
        :return: the number of satellites in the constellation
        :rtype: int
        """
        return self._num_satellites

    @num_satellites.setter
    def num_satellites(self, num_satellites: int):
        """
        set the number of satellites in the constellation
        :param num_satellites: the number of satellites in the constellation
        :type num_satellites: int
        """
        try:
            num_satellites = int(num_satellites)
        except Exception as e:
            raise TypeError
        # ensure the number of satellites is greater than 0
        if num_satellites < 1:
            raise ValueError
        self._num_satellites = num_satellites

    @property
    def semi_major_axis(self) -> float:
        """
        get the semi-major axis of every spacecraft in the constellation
        :return: the semi-major axis of every spacecraft in the constellation
        :rtype: float
        """
        if self._semi_major_axis is None:
            raise ValueError
        return self._semi_major_axis

    @semi_major_axis.setter
    def semi_major_axis(self, semi_major_axis: float):
        """
        set the semi-major axis of every spacecraft in the constellation
        :param semi_major_axis: the semi-major axis of every spacecraft in the constellation
        :type semi_major_axis: float
        """
        try:
            semi_major_axis = float(semi_major_axis)
        except Exception as e:
            raise TypeError
        # ensure the semi-major axis is greater than 0
        if semi_major_axis < 0:
            raise ValueError
        self._semi_major_axis = semi_major_axis

    @property
    def eccentricity(self) -> float:
        """
        get the eccentricity of every spacecraft in the constellation
        :return: the eccentricity of every spacecraft in the constellation
        :rtype: float
        """
        if self._eccentricity is None:
            raise ValueError
        return self._eccentricity

    @eccentricity.setter
    def eccentricity(self, eccentricity: float):
        """
        set the eccentricity of every spacecraft in the constellation
        :param eccentricity: the eccentricity of every spacecraft in the constellation
        :type eccentricity: float
        """
        try:
            eccentricity = float(eccentricity)
        except Exception as e:
            raise TypeError
        # ensure the eccentricity is greater than 0
        if eccentricity < 0:
            raise ValueError
        self._eccentricity = eccentricity

    @property
    def inclination(self) -> float:
        """
        get the inclination of every spacecraft in the constellation
        :return: the inclination of every spacecraft in the constellation [rad]
        :rtype: float
        """
        if self._inclination is None:
            raise ValueError
        return self._inclination

    @inclination.setter
    def inclination(self, inclination: float):
        """
        set the inclination of every spacecraft in the constellation
        :param inclination: the inclination of every spacecraft in the constellation [rad]
        :type inclination: float
        """
        try:
            inclination = float(inclination)
        except Exception as e:
            raise TypeError
        self._inclination = inclination

    @property
    def right_ascension(self) -> float:
        """
        get the right ascension of every spacecraft in the constellation
        :return: the right ascension of every spacecraft in the constellation [rad]
        :rtype: float
        """
        if self._right_ascension is None:
            raise ValueError
        return self._right_ascension

    @right_ascension.setter
    def right_ascension(self, right_ascension: float):
        """
        set the right ascension of every spacecraft in the constellation
        :param right_ascension: the right ascension of every spacecraft in the constellation [rad]
        :type right_ascension: float
        """
        try:
            right_ascension = float(right_ascension)
        except Exception as e:
            raise TypeError
        self._right_ascension = right_ascension

    @property
    def argument_of_periapsis(self) -> float:
        """
        get the argument of periapsis of every spacecraft in the constellation
        :return: the argument of periapsis of every spacecraft in the constellation [rad]
        :rtype: float
        """
        if self._argument_of_periapsis is None:
            raise ValueError
        return self._argument_of_periapsis

    @argument_of_periapsis.setter
    def argument_of_periapsis(self, argument_of_periapsis: float):
        """
        set the argument of periapsis of every spacecraft in the constellation
        :param argument_of_periapsis: the argument of periapsis of every spacecraft in the constellation [rad]
        :type argument_of_periapsis: float
        """
        try:
            argument_of_periapsis = float(argument_of_periapsis)
        except Exception as e:
            raise TypeError
        self._argument_of_periapsis = argument_of_periapsis

    @property
    def true_anomaly_offset(self) -> float:
        """
        get the true anomaly offset of every spacecraft in the constellation
        :return: the true anomaly offset of every spacecraft in the constellation [rad]
        :rtype: float
        """
        if self._true_anomaly_offset is None:
            raise ValueError
        return self._true_anomaly_offset

    @true_anomaly_offset.setter
    def true_anomaly_offset(self, true_anomaly_offset: float):
        """
        set the true anomaly offset of every spacecraft in the constellation
        :param true_anomaly_offset: the true anomaly offset of every spacecraft in the constellation [rad]
        :type true_anomaly_offset: float
        """
        try:
            true_anomaly_offset = float(true_anomaly_offset)
        except Exception as e:
            raise TypeError
        self._true_anomaly_offset = true_anomaly_offset

    @property
    def spacecraft(self) -> dict:
        """
        get the spacecraft in the constellation
        :return: the spacecraft in the constellation
        :rtype: dict
        """
        return self._spacecraft

    @spacecraft.setter
    def spacecraft(self, spacecraft: dict):
        """
        set the spacecraft in the constellation
        :param spacecraft: the spacecraft in the constellation
        :type spacecraft: dict
        """
        if not isinstance(spacecraft, dict):
            raise TypeError
        self._spacecraft = spacecraft

    def __init__(self, init_classical_elements=False, **kwargs):
        """
        initialize the constellation

        :param init_classical_elements: whether to initialize the orbital elements for every spacecraft in the
            constellation
        :type init_classical_elements: bool
        :param num_satellites: the number of satellites in the constellation
        :type num_satellites: int
        :param semi_major_axis: the semi-major axis of every spacecraft in the constellation
        :type semi_major_axis: float
        :param eccentricity: the eccentricity of every spacecraft in the constellation
        :type eccentricity: float
        :param inclination: the inclination of every spacecraft in the constellation
        :type inclination: float
        :param right_ascension: the right ascension of every spacecraft in the constellation
        :type right_ascension: float
        :param argument_of_periapsis: the argument of periapsis of every spacecraft in the constellation
        :type argument_of_periapsis: float
        :param true_anomaly_offset: the true anomaly offset of every spacecraft in the constellation
        :type true_anomaly_offset: float
        :param kwargs:
        """
        super().__init__()
        self.num_satellites = kwargs.get("num_satellites", 1)
        if "semi_major_axis" in kwargs:
            self.semi_major_axis = kwargs.get("semi_major_axis")
        if "eccentricity" in kwargs:
            self.eccentricity = kwargs.get("eccentricity")
        if "inclination" in kwargs:
            self.inclination = kwargs.get("inclination")
        if "right_ascension" in kwargs:
            self.right_ascension = kwargs.get("right_ascension")
        if "argument_of_periapsis" in kwargs:
            self.argument_of_periapsis = kwargs.get("argument_of_periapsis")
        if "true_anomaly_offset" in kwargs:
            self.true_anomaly_offset = kwargs.get("true_anomaly_offset")
        # initialize the spacecraft dictionary
        self.spacecraft = {}
        for i in range(self.num_satellites):
            self.spacecraft[i] = dict()
        # initialize the orbital elements for every spacecraft in the constellation
        if init_classical_elements:
            self.init_classical_elements(**kwargs)

    def set_mean_elements(
        self,
        semi_major_axis_mean: float = None,
        eccentricity_mean: float = None,
        inclination_mean: float = None,
        right_ascension_mean: float = None,
        argument_of_periapsis_mean: float = None,
        true_anomaly_mean: float = None,
        **kwargs,
    ):
        """
        set the mean orbital elements for every spacecraft in the constellation

        :param semi_major_axis_mean: the mean semi-major axis of every spacecraft in the constellation
        :type semi_major_axis_mean: float
        :param eccentricity_mean: the mean eccentricity of every spacecraft in the constellation
        :type eccentricity_mean: float
        :param inclination_mean: the mean inclination of every spacecraft in the constellation
        :type inclination_mean: float
        :param right_ascension_mean: the mean offset in right ascension
        :type right_ascension_mean: float
        :param argument_of_periapsis_mean: the mean offset in argument of periapsis
        :type argument_of_periapsis_mean: float
        :param true_anomaly_mean: the mean offset in true anomaly
        :type true_anomaly_mean: float
        :return:
        """
        if semi_major_axis_mean is None:
            raise ValueError(
                "Cannot set the mean orbital elements if the semi_major_axis is None"
            )
        if eccentricity_mean is None:
            raise ValueError(
                "Cannot set the mean orbital elements if the eccentricity is None"
            )
        if inclination_mean is None:
            raise ValueError(
                "Cannot set the mean orbital elements if the inclination is None"
            )
        if right_ascension_mean is None:
            raise ValueError(
                "Cannot set the mean orbital elements if the right_ascension is None"
            )
        if argument_of_periapsis_mean is None:
            raise ValueError(
                "Cannot set the mean orbital elements if the argument_of_periapsis is None"
            )
        if true_anomaly_mean is None:
            raise ValueError(
                "Cannot set the mean orbital elements if the true_anomaly is None"
            )
        planet = kwargs.get("planet", "earth")
        req = get_planet_property(planet, "REQ")
        j2 = get_planet_property(planet, "J2")
        # convert the mean to their equivalent osculating orbital elements so they can be stored
        elements_osculating = mean_to_osculating_elements(
            req=req,
            j2=j2,
            semi_major_axis=semi_major_axis_mean,
            eccentricity=eccentricity_mean,
            inclination=inclination_mean,
            right_ascension=right_ascension_mean,
            argument_of_periapsis=argument_of_periapsis_mean,
            true_anomaly=true_anomaly_mean,
            mean_to_osculating=True,
        )
        self.semi_major_axis = elements_osculating[0]
        self.eccentricity = elements_osculating[1]
        self.inclination = elements_osculating[2]
        self.right_ascension = elements_osculating[3]
        self.argument_of_periapsis = elements_osculating[4]
        self.true_anomaly_offset = elements_osculating[5]

    @abstractmethod
    def init_classical_elements(self, *args, **kwargs):
        """
        initialize the orbital elements for every spacecraft in the constellation, if the orbital element
            properties of this class are osculating then the initial orbital elements will be osculating, otherwise
            they will be mean
        :return:
        """
        raise NotImplementedError

    @classical_elements_mean(mean_to_osculating=True)
    def init_classical_elements_osculating(self, *args, **kwargs):
        """
        initialize the osculating classical orbital elements for every spacecraft in the constellation assuming that the
            orbital element properties of this object are mean
        :param mean_to_osculating: whether to convert the mean orbital elements to their equivalent osculating orbital
            elements
        :type mean_to_osculating: bool
        :return:
        """
        return self.init_classical_elements(*args, **kwargs)

    @classical_elements_mean(mean_to_osculating=False)
    def init_classical_elements_mean(self, *args, **kwargs):
        """
        initialize the mean classical orbital elements for every spacecraft in the constellation assuming that the
            orbital element properties of this object are osculating
        :return:
        """
        return self.init_classical_elements(*args, **kwargs)

    @state_vectors
    def init_state_vectors(self, *args, **kwargs):
        """
        initialize the inertial state vectors for every spacecraft in the constellation
        :return:
        """
        return self.init_classical_elements()

    @state_vectors
    def init_state_vectors_mean(self, *args, **kwargs):
        """
        initialize the mean inertial state vectors for every spacecraft in the constellation
        :return:
        """
        return self.init_classical_elements_mean(*args, **kwargs)

    @state_vectors
    def init_state_vectors_osculating(self, *args, **kwargs):
        """
        initialize the osculating inertial state vectors for every spacecraft in the constellation
        :return:
        """
        return self.init_classical_elements_osculating(*args, **kwargs)

    def __iter__(self):
        """
        iterate over the spacecraft in the constellation
        :return: an iterator over the spacecraft in the constellation
        """
        return iter(self._spacecraft.items())

    def __getitem__(self, item):
        """
        get a spacecraft in the constellation
        :param item: the spacecraft id
        :return: the spacecraft
        """
        return self._spacecraft[item]

    def __setitem__(self, key, value):
        """
        set a spacecraft in the constellation
        :param key: the spacecraft id
        :param value: the spacecraft
        """
        self._spacecraft[key] = value

    def __delitem__(self, key):
        """
        delete a spacecraft in the constellation
        :param key: the spacecraft id
        """
        del self._spacecraft[key]

    def __len__(self):
        """
        get the number of spacecraft in the constellation
        :return: the number of spacecraft in the constellation
        """
        return len(self._spacecraft)

    def __contains__(self, item):
        """
        check if a spacecraft is in the constellation
        :param item: the spacecraft id
        :return: true if the spacecraft is in the constellation, false otherwise
        """
        return item in self._spacecraft

    def __str__(self):
        """
        get the string representation of the constellation
        :return: the string representation of the constellation
        """
        return str(self._spacecraft)

    def __repr__(self):
        """
        get the string representation of the constellation
        :return: the string representation of the constellation
        """
        return repr(self._spacecraft)

    def __eq__(self, other):
        """
        check if two constellations are equal
        :param other: the other constellation
        :return: true if the two constellations are equal, false otherwise
        """
        raise NotImplementedError

    def __ne__(self, other):
        """
        check if two constellations are not equal
        :param other: the other constellation
        :return: true if the two constellations are not equal, false otherwise
        """
        return not self.__eq__(other)

    def __hash__(self):
        """
        get the hash of the constellation
        :return: the hash of the constellation
        """
        return hash(self._spacecraft)

    def __bool__(self):
        """
        check if the constellation is not empty
        :return: true if the constellation is not empty, false otherwise
        """
        return bool(self._spacecraft)

    def __enter__(self):
        """
        enter the context of the constellation
        :return: the constellation
        """
        self.init_classical_elements()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        exit the context of the constellation
        :param exc_type: the exception type
        :param exc_val: the exception value
        :param exc_tb: the exception traceback
        """
        pass

    def iter_state_vectors(self, *args, **kwargs):
        """
        iterate over the state vectors of the spacecraft in the constellation
        :return: an iterator over the state vectors of the spacecraft in the constellation
        """
        return iter(self.init_state_vectors(*args, **kwargs).items())

    def true_argument_of_latitude(self, spacecraft_id: int):
        """
        get the true argument of latitude of the spacecraft in the constellation
        :return: the true argument of latitude of the spacecraft in the constellation
        """
        return argument_of_latitude(
            argument_of_periapsis=self[spacecraft_id][4],
            anomaly=self[spacecraft_id][5],
        )

    def is_valid_spacecraft_id(self, spacecraft_id: int) -> bool:
        """
        check if the spacecraft id is valid
        :param spacecraft_id: the spacecraft id
        :return: true if the spacecraft id is valid, false otherwise
        """
        return spacecraft_id in self.spacecraft.keys()

    def set_variable(
        self, spacecraft_ids: Optional[Union[int, List[int]]] = None, **kwargs
    ):
        """
        set variables for spacecraft in the constellation where the variables are input into the method as keyword
            arguments
        :param spacecraft_ids: the spacecraft ids to set the variable for. If None, then the variable will be set for
            every spacecraft in the constellation
        :type spacecraft_ids: Optional[Union[int, List[int]]]
        :param kwargs: the variables to set for the spacecraft in the constellation
        """
        # if the spacecraft ids are None, then set the variable for every spacecraft in the constellation
        if spacecraft_ids is None:
            spacecraft_ids = range(len(self))
        # if the spacecraft ids are an integer, then convert it to a list
        if isinstance(spacecraft_ids, int):
            spacecraft_ids = [spacecraft_ids]
        # set the variable for every spacecraft in the constellation
        for spacecraft_id in spacecraft_ids:
            for variable, value in kwargs.items():
                if not self.is_valid_spacecraft_id(spacecraft_id):
                    raise KeyError(f"The spacecraft id {spacecraft_id} does not exist")

                # if the value is a container, then set the variable for every spacecraft in the constellation
                if isinstance(value, Container) and not isinstance(value, str):
                    if len(value) != len(spacecraft_ids):
                        raise ValueError(
                            f"The length of the value {variable} does not match the number of "
                            + f"spacecraft in the constellation {len(self)}"
                        )
                    self[spacecraft_id][variable] = value[spacecraft_id]
                else:
                    self[spacecraft_id][variable] = value


class Coplanar(Constellation):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def init_classical_elements(self, *args, **kwargs):
        """
        initialize the orbital elements for every spacecraft in the constellation
        :return:
        """
        relative_phase: float = 2 * np.pi / self.num_satellites
        # use dictionary comprehension to efficiently initialize the orbital elements for every spacecraft in the
        #   constellation
        for i in range(self.num_satellites):
            self.spacecraft[i]["semi_major_axis"] = self.semi_major_axis
            self.spacecraft[i]["eccentricity"] = self.eccentricity
            self.spacecraft[i]["inclination"] = self.inclination
            self.spacecraft[i]["right_ascension"] = self.right_ascension
            self.spacecraft[i]["argument_of_periapsis"] = self.argument_of_periapsis
            self.spacecraft[i]["true_anomaly"] = (
                self.true_anomaly_offset + i * relative_phase
            )
        return self._spacecraft


class CoplanarCircular(Coplanar):

    _eccentricity = 0.0

    @property
    def eccentricity(self) -> float:
        return 0.0

    @eccentricity.setter
    def eccentricity(self, eccentricity: float):
        self._eccentricity = 0.0

    def __init__(self, **kwargs):
        self.eccentricity = 0
        super().__init__(**kwargs)


class Walker(Constellation):
    """
    A Walker constellation is a constellation of satellites that are evenly spaced in a Walker pattern.
    """

    _num_planes: int = 1

    @property
    def num_planes(self) -> int:
        """
        get the number of planes in the constellation
        :return: the number of planes in the constellation
        :rtype: int
        """
        return self._num_planes

    @num_planes.setter
    def num_planes(self, num_planes: int):
        """
        set the number of planes in the constellation
        :param num_planes: the number of planes in the constellation
        :type num_planes: int
        """
        try:
            num_planes = int(num_planes)
        except Exception as e:
            raise TypeError
        # ensure the number of planes is greater than 0
        if num_planes < 1:
            raise ValueError
        self._num_planes = num_planes

    def __init__(self, **kwargs):
        self.num_planes = kwargs.get("num_planes", 1)
        super().__init__(**kwargs)


class WalkerDelta(Walker):
    """
    A Walker Delta constellation is a constellation of satellites that are evenly spaced in a Walker Delta pattern.
    """

    _relative_spacing = 1.0

    @property
    def relative_spacing(self) -> float:
        """
        get the relative spacing of the constellation
        :return: the relative spacing of the constellation
        :rtype: float
        """
        return self._relative_spacing

    @relative_spacing.setter
    def relative_spacing(self, relative_spacing: float):
        """
        set the relative spacing of the constellation
        :param relative_spacing: the relative spacing of the constellation
        :type relative_spacing: float
        """
        try:
            relative_spacing = float(relative_spacing)
        except Exception as e:
            raise TypeError
        # ensure the relative spacing is greater than or equal to 0
        if relative_spacing < 0:
            raise ValueError
        self._relative_spacing = relative_spacing

    def __init__(self, **kwargs):
        self.relative_spacing = kwargs.get("relative_spacing", 1.0)
        super().__init__(**kwargs)

    def init_classical_elements(self, *args, **kwargs):
        """
        initialize the orbital elements for every spacecraft in the constellation
        :param args:
        :param kwargs:
        :return:
        """
        if self.num_planes > self.num_satellites:
            self.num_planes = self.num_satellites
        # define the full rotation
        full_rotation = 2 * np.pi
        # get the relative spacing parameter which must be within 0 and the number of planes
        relative_spacing = min(max(abs(self.relative_spacing), 0), self.num_planes)
        # calculate the relative phase which will determine the phase offset in true anomaly of spacecraft in every
        #   plane
        relative_phase = relative_spacing * full_rotation / self.num_satellites
        # perform integer division to get the number of satellites per plane as a discrete number
        num_satellites_plane = self.num_satellites // self.num_planes
        # calculate the relative anomaly which will determine the difference in true anomaly between every spacecraft
        relative_anom = full_rotation / num_satellites_plane
        # raise an exception if the number of satellites isn't perfectly divisible into the number of planes
        if self.num_satellites % self.num_planes != 0:
            raise ValueError(
                "The number of satellites should be divisible by the number of planes"
            )
        # set the initial orbital elements for each spacecraft in the constellation
        for i in range(self.num_planes):
            for j in range(num_satellites_plane):
                true_anom = (
                    relative_phase * i + relative_anom * j + self.true_anomaly_offset
                )
                raan = full_rotation / self.num_planes * i + self.right_ascension
                index = i * num_satellites_plane + j
                self.spacecraft[index]["semi_major_axis"] = self.semi_major_axis
                self.spacecraft[index]["eccentricity"] = self.eccentricity
                self.spacecraft[index]["inclination"] = self.inclination
                self.spacecraft[index]["right_ascension"] = raan
                self.spacecraft[index]["argument_of_periapsis"] = self.argument_of_periapsis
                self.spacecraft[index]["true_anomaly"] = true_anom
        # return the orbital elements
        return self._spacecraft



#                     [ ZENDIR ]
# This code is developed by Zendir to aid with communication
# to the public API. All code is under the the license provided
# with the 'zendir' module. Copyright Zendir, 2025.

from ..utils import printer, ZendirException, helper
from .instance import Instance
from .message import Message
from .context import Context


class Behaviour(Instance):
    """
    The Behaviour class can define an instance that can exist within the simulation.
    Behaviours can be added to both the simulation or to an object that exists. They
    have no physical representation (no position or rotation) and purely exist as a
    way to serve abstract functionality. Examples of behaviours include flight software
    and power modules.
    """

    __parent: Instance = None
    """Defines the parent object that the behaviour is attached to."""

    __messages: dict[str:Message] = {}
    """Defines all messages that are attached to the object, by name."""

    def __init__(
        self, context: Context, id: str, type: str = None, parent: Instance = None
    ) -> None:
        """
        Initialises the behaviour with the context and the ID of the behaviour.

        :param context:         The context of the simulation
        :type context:          Context
        :param id:              The GUID ID of the behaviour
        :type id:               str
        :param type:            The type of the system, if applicable
        :type type:             str
        :param parent:          The parent object that the behaviour is attached to
        :type parent:           Instance
        """

        super().__init__(context, id, type)

        # Clear and reset any data
        self.__parent = parent
        self.__messages = {}

    def get_parent(self) -> Instance:
        """
        Returns the parent object that the behaviour is attached to.

        :returns:   The parent object that the behaviour is attached to
        :rtype:     Object
        """

        return self.__parent

    def _reset_refresh_cache(self) -> None:
        """
        Overrides the base class method to set the flag for refreshing the cache to true.
        This will ensure that all messages will also require a refresh.
        """

        # Ensure all messages require a refresh too
        for message in self.__messages.values():
            message._reset_refresh_cache()
        super()._reset_refresh_cache()

    def get_instance_with_id(self, id: str, recurse: bool = False) -> Instance:
        """
        Returns the instance that is attached to this behaviour with the specified ID. If the
        instance does not exist, None will be returned. This may be a message or a child instance
        of some kind.

        :param id:  The ID of the instance to fetch
        :type id:   str
        :param recurse:  Whether to look down the chain of children to find the instance
        :type recurse:   bool

        :returns:   The instance that is attached to the object with the specified ID
        :rtype:     Instance
        """

        # Start by looking at the instances for the ID
        for msg in self.__messages.values():
            if msg.get_id() == id:
                return msg

        # Return none otherwise
        return None

    async def get_message(self, name: str) -> Message:
        """
        Attempts to get the message with the specified name that is attached to the behaviour. If the
        message does not exist, it will be created and attached to the behaviour. If the message cannot
        be created, an exception will be raised.

        :param name:    The name of the message to fetch
        :type name:     str

        :returns:       The message with the specified name that is attached to the behaviour
        :rtype:         Message
        """

        # Check if the name is within the message structure and return that
        if name in self.__messages.keys():
            return self.__messages[name]

        # Fetch the data
        message_id: str = await self.get(name)
        if not helper.is_valid_guid(message_id):
            raise ZendirException(f"Failed to find message with name '{name}'.")

        # Create the message object with the ID
        message = Message(self._context, message_id)
        self.__messages[name] = message

        # Return the message of that name
        printer.success(f"Successfully created message with name '{name}'.")
        return message

    def get_messages(self) -> list[Message]:
        """
        Returns all of the messages that are attached to the behaviour. This will only include the
        messages that have currently been fetched.

        :returns:   All of the messages that are attached to the behaviour
        :rtype:     list[Message]
        """

        return list(self.__messages.values())
